/*
* All questions regarding the software should be addressed to
*
*       Prof. Charles A. Bouman
*       Purdue University
*       School of Electrical and Computer Engineering
*       1285 Electrical Engineering Building
*       West Lafayette, IN 47907-1285
*       USA
*       +1 765 494 0340
*       +1 765 494 3358 (fax)
*       email:  bouman@ecn.purdue.edu
*       http://www.ece.purdue.edu/~bouman
*
* Copyright (c) 1995 The Board of Trustees of Purdue University.
*
* Permission to use, copy, modify, and distribute this software and its
* documentation for any purpose, without fee, and without written agreement is
* hereby granted, provided that the above copyright notice and the following
* two paragraphs appear in all copies of this software.
*
* IN NO EVENT SHALL PURDUE UNIVERSITY BE LIABLE TO ANY PARTY FOR DIRECT,
* INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
* USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF PURDUE UNIVERSITY HAS
* BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
* PURDUE UNIVERSITY SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
* PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS,
* AND PURDUE UNIVERSITY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
* UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "smap_defs.h"
#include "smap_region.h"
#include "smap_io.h"
#include "smap_decimate.h"
#include "smap_interp.h"
#include "clust_defs.h"
#include "clust_io.h"
#include "clust_util.h"
#include "allocate.h"
#include "alloc_util.h"

void extract_init(struct SigSet *S);
void extract(
  CELL ***img,             /* multispectral image, img[band][i][j] */
  struct Region *region,   /* region to extract */
  LIKELIHOOD ***ll,        /* log likelihood, ll[i][j][class] */
  struct SigSet *S         /* class signatures */
);


static void read_block(
  CELL ***img,
  int  wd,
  int  ht,
  int  nbands,
  struct Region *region,
  char *infn
);
static void shift_ll(
  LIKELIHOOD ****ll_pym,
  struct Region *region,
  int block_size
);
static void shift_img(
  CELL ***img,
  int nbands,
  struct Region *region,
  int block_size
);
static int increment_reg(
  struct Region *region,
  int wd,
  int ht,
  int block_size
);
static void init_reg(
  struct Region *region,
  int wd,
  int ht,
  int block_size
);
static void error(char *name);


int main(argc,argv)
int     argc;
char    *argv[];
{
  int argpt;
  char   *optarg;
  signed char   c;
 
  FILE *fd;
  CELL ***img;                /* multispectral image, img[band][i][j] */
  int  wd,ht;                 /* image width and height */
  struct Region region;       /* specifies image subregion */
  int block_size;             /* size of subregion blocks */
  int nbands;                 /* number of bands */
  int nclasses;               /* number of classes */
  int j;                      /* indexes for band and column */
  LIKELIHOOD ****ll_pym;      /* pyramid of log likelihoods */
  IMAGE_CLASS ***sf_pym;      /* pyramid of segmentations */
  int  D;                     /* number of levels in pyramid */
  struct SigSet S;            /* class parameters */
  double *alpha_dec;          /* class transition probabilities */
  char *infn;                 /* input image file name */
  char *parfn;                /* input parameter set file name, and number */
  char *otfn;                 /* output segmentation file name */
  int  vlevel;                /* level of verbose output */
  int  iflag,pflag,oflag,xflag,yflag,MLflag,SMAPflag,Cflag;
  int  i;

  /* malloc_debug(2); */

  vlevel=1;
  block_size = 128;
  iflag=pflag=oflag=xflag=yflag=MLflag=SMAPflag=Cflag=0;

  argpt = 0;
  while(argpt<argc) {
    if( argv[argpt][0] == '-' ) {
      c = argv[argpt][1];
      if( (argpt+1)<argc ) optarg = argv[argpt+1];
      else optarg = NULL;

      switch(c) {
        case 'h':
          error(argv[0]);
          break;
        case 'i':
          iflag=1;
          infn = (char *)G_malloc((strlen(optarg)+1)*sizeof(char));
          strcpy(infn,optarg);
          break;
        case 'p':
          pflag=1;
          parfn = (char *)G_malloc((strlen(optarg)+1)*sizeof(char));
          strcpy(parfn,optarg);
          break;
        case 'o':
          oflag=1;
          otfn = (char *)G_malloc((strlen(optarg)+1)*sizeof(char));
          strcpy(otfn,optarg);
          break;
        case 'x':
          xflag = 1;
          if(0==sscanf(optarg,"%d",&wd)) error(argv[0]);
          if(wd<1) error(argv[0]);
          break;
        case 'y':
          yflag = 1;
          if(0==sscanf(optarg,"%d",&ht)) error(argv[0]);
          if(ht<1) error(argv[0]);
          break; 
        case 'v':
          if(0==sscanf(optarg,"%d",&vlevel)) error(argv[0]);
          if(vlevel<0) error(argv[0]);
          break;
        case 'M':
          MLflag = 1;
          break;
        case 'S':
          SMAPflag = 1;
          break;
        case 'C':
          Cflag = 1;
          break;
        case 'B':
          if(0==sscanf(optarg,"%d",&block_size)) error(argv[0]);
          if(block_size<8) error(argv[0]);
          for(i=0; (block_size>>i)>1; i++);
          block_size = 1<<i;
          break;
        default:
          error(argv[0]);
          break;
      }
    }
    argpt++;
  }

  /* Check for inconsistant inputs */   
  if(!(xflag&&yflag&&oflag&&iflag&&pflag)) {
    fprintf(stderr, "\n%s: Input, signatures, output, width and height required!\n",argv[0]);
    error(argv[0]);
  }
  if(MLflag&&SMAPflag) {
    fprintf(stderr, "\n%s: Maxlike and Smap can not both be selected!\n",argv[0]);
    error(argv[0]);
  }

  /* set default */
  if((!MLflag)&&(!SMAPflag)) SMAPflag = 1;

  /* read signatures */
  fd = fopen (parfn, "r");
    if (fd == NULL) {
        fprintf(stderr,"\nError: can not open signature file %s\n",parfn);
        exit(10);
    }
  I_ReadSigSet(fd, &S);
  nbands = S.nbands;
  nclasses = S.nclasses;
  if(vlevel>=3) I_WriteSigSet(stdout, &S);

  /* Check for too many classes */
  if(nclasses>MAX_IMAGE_CLASS) {
    fprintf(stderr, "number of classes must be <= %d\n", MAX_IMAGE_CLASS);
    exit(1);
  }

  /* allocate alpha_dec parameters */
  D = levels(block_size,block_size);
  alpha_dec = (double *)G_malloc(D*sizeof(double));

  /* allocate image block */
  img = (CELL ***)multialloc(sizeof(CELL),3,nbands,block_size,block_size);

  /* allocate memory for log likelihood pyramid */
  ll_pym = (LIKELIHOOD ****)get_cubic_pyramid(block_size,block_size,nclasses,sizeof(LIKELIHOOD));

  /* allocate memory for segmentation pyramid */
  sf_pym = (IMAGE_CLASS ***)get_pyramid(wd,ht,sizeof(IMAGE_CLASS));

  /* tiled segmentation */
  init_reg(&region,wd,ht,block_size);
  extract_init(&S);
  do {
    if(vlevel>=1) 
      printf("Segmenting region xmin= %d, ymin= %d\n",region.xmin,region.ymin);
    shift_img(img,nbands,&region,block_size);
    read_block(img,wd,ht,nbands,&region,infn);

    shift_ll(ll_pym,&region,block_size);
    extract(img,&region,ll_pym[0],&S);

    if(SMAPflag) {
      for(i=0; i<D; i++) alpha_dec[i]=1.0;
      seq_MAP(sf_pym,&region,ll_pym,nclasses,alpha_dec,vlevel);
    }

    if(MLflag)
      MLE(sf_pym[0],ll_pym[0],&region,nclasses);

  } while(increment_reg(&region,wd,ht,block_size));

  /* convert classes */
  if(Cflag) {
    for(i=0; i<ht; i++)
    for(j=0; j<wd; j++)
     sf_pym[0][i][j] = S.ClassSig[sf_pym[0][i][j]].classnum;
  }
  else {
    /* increment output for grass GIS program */
    for(i=0; i<ht; i++)
    for(j=0; j<wd; j++) ++(sf_pym[0][i][j]);
  }

  /* write out results */
  write_img(sf_pym[0],wd,ht,otfn);

  return(0);
}


static void init_reg(
  struct Region *region,
  int wd,
  int ht,
  int block_size
)
{
   region->xmin = 0;
   region->ymin = 0;

   region->xmax = block_size;
   if(region->xmax>wd)
     region->xmax=wd;

   region->ymax = block_size;
   if(region->ymax>ht) 
     region->ymax=ht;

   region->free.left = 1;
   region->free.top = 1;
   region->free.right = 1;
   region->free.bottom = 1;
}


static int increment_reg(
  struct Region *region,
  int wd,
  int ht,
  int block_size
)
{

   /* shift one block to right */
   if(region->xmax<wd) {
     region->xmin = region->xmax;
     region->xmax = region->xmin + block_size;
     if(region->xmax>wd) region->xmax = wd;
   }
   else {
     /* shift one block down */
     if(region->ymax<ht) {
       region->xmin = 0;
       region->xmax = region->xmin + block_size;
       if(region->xmax>wd) region->xmax = wd;

       region->ymin = region->ymax;
       region->ymax = region->ymax + block_size;
       if(region->ymax>ht) region->ymax = ht;
     }
     /* finished shifting */
     else {
       return(0);
     }
   }

   if(region->xmin == 0) region->free.left = 1;
   else region->free.left = 0;

   if(region->ymin == 0) region->free.top = 1;
   else region->free.top = 0;

   region->free.right = 1;
   region->free.bottom = 1;

   return(1);
}


static void shift_img(
  CELL ***img,
  int nbands,
  struct Region *region,
  int block_size
)
{
    static int xoffset=0;
    static int yoffset=0;
    int xdelta;
    int ystart,ystop,ydelta;
    int b,i;

    xdelta = region->xmin - xoffset;
    ydelta = region->ymin - yoffset;
    xoffset = region->xmin;
    yoffset = region->ymin;

    ystart = region->ymin;
    ystop  = ystart + block_size;

    for(b=0; b<nbands; b++) {
      img[b] -= ydelta;
      for(i=ystart; i<ystop; i++ )
        img[b][i] -= xdelta;
    }

}

static void shift_ll(
  LIKELIHOOD ****ll_pym,
  struct Region *region,
  int block_size
)
{
    static int first=1;
    static int xoffset[20];
    static int yoffset[20];

    int xdelta;
    int ystart,ystop,ydelta;
    int D;
    int k,i;
    int block_size_k;
    struct Region region_buff;

    /* if first time, set offsets to 0 */
    if(first) {
      D=levels(block_size,block_size);
      for(k=0; k<=D; k++) 
        xoffset[k]=yoffset[k]=0;
      first=0;
    }

    /* save region information */
    copy_reg(region,&region_buff);

    /* subtract pointer offsets for each scale */
    D=levels(block_size,block_size);
    block_size_k = block_size;
    for(k=0; k<=D; k++) {
      xdelta = region->xmin - xoffset[k];
      ydelta = region->ymin - yoffset[k];
      xoffset[k] = region->xmin;
      yoffset[k] = region->ymin;

      ystart = region->ymin;
      ystop  = ystart + block_size_k;

      ll_pym[k] -= ydelta;
      for(i=ystart; i<ystop; i++ )
        ll_pym[k][i] -= xdelta;
      dec_reg(region);
      block_size_k = block_size_k/2;
    }

    /* replace region information */
    copy_reg(&region_buff,region);
}


static void read_block(
  CELL ***img,
  int  wd,
  int  ht,
  int  nbands,
  struct Region *region,
  char *infn
)
{
     static int first=1;
     static unsigned char ***img_buf;

     int b,i,j;
     int xstop,ystop;
     char *infn_num;

     /* allocate image buffer first time called */
     if(first) {
       img_buf = (unsigned char ***)multialloc(sizeof(unsigned char),3,nbands,ht,wd);

       /* allocate memory for name extension */
       infn_num = (char *)G_malloc((strlen(infn)+10)*sizeof(char));

       /* read data */
       if(nbands==1) {
         read_img(img_buf[0],wd,ht,infn);
       } 
       else { 
         for(b=1; b<=nbands; b++) {
           sprintf(infn_num,"%s.%d",infn,b);
           read_img(img_buf[b-1],wd,ht,infn_num);
         } 
       }
       first=0;
     }

     /* copy data from buffer */
     xstop = region->xmax; if(xstop>wd) xstop=wd;
     ystop = region->ymax; if(ystop>ht) ystop=ht;
     for(b=0; b<nbands; b++) 
     for(i=region->ymin; i<ystop; i++)
     for(j=region->xmin; j<xstop; j++)
       img[b][i][j] = img_buf[b][i][j];
}


static void error(char *name)
{
  fprintf(stderr,"\nUsage: %s\n",name);
  fprintf(stderr,"     -help -- produce this help menu\n");
  fprintf(stderr,"     -i image -- pgm format image(s);\n");
  fprintf(stderr,"        May either be a single image, or may be numbered\n");
  fprintf(stderr,"        numbered image.1 image.2 etc.\n");
  fprintf(stderr,"     -x width  -- Input image width\n");
  fprintf(stderr,"     -y height -- Input image height\n");
  fprintf(stderr,"     -p params -- File containing spectral parameters.\n");
  fprintf(stderr,"     -o output -- Segmentation image in pgm format\n");
  fprintf(stderr,"     -v level  -- Verbose operation; level>=0\n");
  fprintf(stderr,"     -<algorithm> -- May be one of SMAP, Maxlike; Default SMAP\n");
  fprintf(stderr,"     -B blocksize -- Processing block size; Default 128\n");
  fprintf(stderr,"     -Classnum -- Output class number in stead of class index\n");
  exit(1);
}


